(in-package :cl-user)

(defgeneric (setf form) (obj val))

(defclass eve () (form)
 (:documentation "
(form-eve ('eve))
Eve's class (hopefully) encloses a lexically scoped GENSYM
as the name of the class. Once there is no way
to reach that GENSYM, the class is GCable.
'MAKE-LOAD-FORM has a method for EVE
(based on the #'FORM-EVE macro expansion)
Children of an eve should implement 'MAKE-LOAD-FORM
to post-process the EVE created through
(eval (call-next-method))
"))

(defmethod make-load-form 
 ((obj eve) &optional env
  &aux (class (class-of obj)))
 (declare (ignorable obj env))
 (macroexpand-1 
     `(form-eve (,@(mapcar (lambda (x) (list 'quote (class-name `,x)))
                 (c2mop:class-direct-superclasses class))))))

(defmacro form-eve ((&rest other-classes) &rest keys &key &allow-other-keys)
 `(let
   ((form
     `(macrolet
       ((ana-quine-class
         ((&rest class-paraphernalia) &body body)
         (alexandria:with-gensyms (classname)
          `(let ((quine
                  (defclass ,classname
                   ,@class-paraphernalia)))
            ,@body))))
       (ana-quine-class
        ((,,@other-classes)
          ((form :accessor form))
         (:documentation " . . . "))
        (values (make-instance quine ,,@keys))))))
   (let ((particular-quine (eval form)))
     (setf (form particular-quine) form)
     (values particular-quine))))

(defclass lamb (eve)
 ((form-of :initarg :form-of :accessor form-of)
  (func-of :initform nil :reader func-of))
 (:default-initargs :form-of '(lambda ()))
 (:documentation "
LAMB provides two slots:
(form-of '(some form))
and
(func-of (obj lamb))
which is a reader for the 'FUNC-OF slot
which is the 'FORM-OF evaluated.
Double quote forms that aren't self-evaluating.
''
"))

(defmethod (setf form-of) :after (val (obj lamb))
 (setf (slot-value obj 'FUNC-OF)
  (eval (form-of obj))))

(defmethod shared-initialize :after ((obj lamb) names &rest rest)
 (declare (ignore names rest))
 (setf (form-of obj) (form-of obj)))

(defmethod make-load-form 
 ((obj lamb) &optional env
  &aux (class (class-of obj)))
 (declare (ignorable obj env))
 (let ((form (call-next-method)))
  (nconc (caddar (last (car (last (car (cdaadr form))))))
   		;And that's where you'll meet the wizard.
   `(:form-of '',(form-of obj)))
  (values form)))

(defclass mempubsub (eve)
 ((peers :initform (tg:make-weak-hash-table :weakness :value)
   :allocation :class :reader peers))
 (:documentation "
This class only has a meaning for in-memory classes/class 
instances, so it needs no special 'make-load-form.
(bread-massage *mempubsub* #'func)
Iterates over mempubsub class peers, and calls func on members
*mempubsub*'s class is a direct superclass of.
(push #+myclass (c2mop:class-direct-superclasses *adoptee*))
is how you subscribe if not
(form-eve ((class-name (class-of *mempubsub*))))
"))

(defmethod shared-initialize :after ((obj mempubsub) names
                                     &rest rest)
 (declare (ignore names rest))
 (setf (gethash obj (peers obj)) obj))

(defmethod spread-message ((obj mempubsub) (func function))
 (with-hash-table-iterator (iter (peers obj))
  (loop initially (tg:gc)
       for k = (nth-value 1 (iter)) while k
       when (and (member (class-of obj)
                  (c2mop:class-direct-superclasses
                   (class-of k))))
           do (funcall func k))))

(defclass cloning (eve) ())

(defmethod make-load-form
 ((obj cloning) &optional env
  &aux (class (class-of obj))) "
"
 (declare (ignore env))
 (alexandria:with-gensyms (form classer)
 `(let ((,form ,(call-next-method)))
   (change-class (eval ,form) ,class))))

